#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import os

import unittest
from taurus.test import insertTest

from sardana.macroserver.macroserver import MacroServer
from sardana.macroserver.scan.recorder import DataRecorder
from sardana.macroserver.scan.recorder.storage import FileRecorder,\
    BaseFileRecorder


_TEST_DIR = os.path.dirname(os.path.abspath(__file__))
_FAKE_RECORDER_DIR = os.path.join(_TEST_DIR, 'res')


@insertTest(helper_name='getRecorderClass', klass_name="JsonRecorder")
@insertTest(helper_name='getRecorderClass', klass_name="FIO_FileRecorder")
@insertTest(helper_name='getRecorderClass', klass_name="FakeScanRecorder",
            extra_paths=[_FAKE_RECORDER_DIR])
@insertTest(helper_name='getRecorderClasses', filter=DataRecorder,
            extra_paths=[_FAKE_RECORDER_DIR], extra_recorders=1)
@insertTest(helper_name='getRecorderClasses', extra_paths=[_FAKE_RECORDER_DIR],
            extra_recorders=1)
@insertTest(helper_name='getRecorderPath',
            recorder_path=["/tmp/foo", "#/tmp/foo2"], expected_num_path=2)
@insertTest(helper_name='getRecorderPath', recorder_path=["/tmp/foo:/tmp/foo2"],
            expected_num_path=3)
@insertTest(helper_name='getRecorderPath', recorder_path=["/tmp/foo"],
            expected_num_path=2)
@insertTest(helper_name='getRecorderPath')
class RecorderManagerTest(unittest.TestCase):
    # Just an hardcode fullname for create an instance of MacroServer.
    # This macroserver does not need to be defined.
    ms_fullname = "macroserver/demo1/1"

    def setUp(self):
        name = self.ms_fullname.split("/")[1]
        self._macro_server = MacroServer(self.ms_fullname, name, macro_path=[],
                                         recorder_path=[])
        self.manager = self._macro_server.recorder_manager

    def tearDown(self):
        pass

    def _updateRecorderManager(self, recorder_path):
        """Helper for update the sardana recorder manager
        """
        self.manager.setRecorderPath(recorder_path)

    def getRecorderPath(self, recorder_path=[], expected_num_path=1):
        """Helper for test the number of reading recorder paths.
        The number of reading path sould be len(recorder_path) + 1
        """
        if recorder_path is not []:
            self._updateRecorderManager(recorder_path)
        # Get the list of recorder path(s)
        paths = self.manager.getRecorderPath()
        num_paths = len(paths)
        msg = "The number of paths do not concur, read %d, expected %d" %\
              (num_paths, expected_num_path)
        self.assertEqual(num_paths, expected_num_path, msg)

    def getRecorderClasses(self, filter=None, extra_paths=None,
                           extra_recorders=0):
        """Helper for test getRecorderClasses method of the record Manager.
        """
        if filter is None:
            filter = DataRecorder
        # Use default recorders paths
        self.manager.setRecorderPath([])
        default_recorder_klass = self.manager.getRecorderClasses(filter)
        # Add extra recorders paths
        if extra_paths is not None:
            self.manager.setRecorderPath(extra_paths)
        recorder_klass = self.manager.getRecorderClasses(filter)
        n_default_recorders = len(default_recorder_klass)
        n_recorders = len(recorder_klass)
        total_recorders = n_default_recorders + extra_recorders
        msg = "Number of recorder classes do not concur, expected %d, get %d" %\
              (total_recorders, n_recorders)
        self.assertEqual(total_recorders, n_recorders, msg)

    def getRecorderClass(self, klass_name, extra_paths=[]):
        """Helper for test getRecorderClass method of the record Manager.
        """
        self.manager.setRecorderPath(extra_paths)
        klass = self.manager.getRecorderClass(klass_name)
        msg = "Recoder manager does not found the class %s" % (klass_name)
        self.assertNotEqual(klass, None, msg)
        _name = klass.__name__
        msg = "The class %s is not subclass of DataRecorder" % (_name)
        self.assertTrue(issubclass(klass, DataRecorder), msg)
        msg = "The class name giveb by the recorder manager is different." +\
              "Expected %s, get %s" % (klass_name, _name)
        self.assertEqual(_name, klass_name, msg)

    def test_SameClassNames(self):
        """Test whether ordered path precedence is maintained in case of
        different recorder classes with the same name located in different
        paths.
        """
        path1 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path1')
        path2 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path2')
        path3 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path3')
        # set three paths containing recorders with the same class names
        recorder_path = [path3, path1, path2]
        self._updateRecorderManager(recorder_path)
        klass = self.manager.getRecorderMetaClass('FakeScanRecorder')
        # retrieve path to the recorder library
        path = os.sep.join(klass.lib.full_name.split(os.sep)[:-1])
        msg = 'Ordered path precedence is not maintained by RecorderManager'
        self.assertEqual(path3, path, msg)

    def test_SameFormats(self):
        """Test whether ordered path precedence is maintained in case of
        different recorder classes supporting the same format located in
        different paths.
        """
        path1 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path1')
        path2 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path2')
        path3 = os.path.join(_TEST_DIR, 'res', 'recorders', 'path3')
        recorder_path = [path3, path1, path2]
        # set three paths containing recorders of the same format
        self._updateRecorderManager(recorder_path)
        klasses = self.manager.getRecorderMetaClasses(filter=BaseFileRecorder,
                                                      extension='.spec')
        klass = list(klasses.values())[0]
        # retrieve path to the recorder library
        path = os.sep.join(klass.lib.full_name.split(os.sep)[:-1])
        msg = 'Ordered path precedence is not maintained by RecorderManager'
        self.assertEqual(path3, path, msg)

    def test_ExternalVsBuiltinPrecedence(self):
        """Test if external recorders are of higher priority than the built-in)
        """
        external_path = os.path.join(_TEST_DIR, 'res', 'recorders',
                                     'pathexternal')

        # set three paths containing recorders with the same class names
        recorder_path = [external_path]
        self._updateRecorderManager(recorder_path)
        klass = self.manager.getRecorderMetaClass('SPEC_FileRecorder')
        # retrieve path to the recorder library
        path = os.sep.join(klass.lib.full_name.split(os.sep)[:-1])
        msg = 'Wrong precedence of recorder paths'
        self.assertEqual(path, external_path, msg)



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import random

from sardana import State
from sardana.pool.controller import ZeroDController


class Channel:

    def __init__(self, idx):
        self.idx = idx            # 1 based index
        self.value = 0.0
        self.active = False


class DummyZeroDController(ZeroDController):
    """This class represents a dummy Sardana 0D controller."""

    gender = "Simulation"
    model = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024

    def __init__(self, inst, props, *args, **kwargs):
        ZeroDController.__init__(self, inst, props, *args, **kwargs)

        self.channels = [Channel(i + 1) for i in range(self.MaxDevice)]
        self.read_channels = {}

    def AddDevice(self, ind):
        self.channels[ind].active = True

    def DeleteDevice(self, ind):
        self.channels[ind].active = False

    def StateOne(self, ind):
        return State.On, "OK"

    def _setChannelValue(self, channel):
        channel.value = 100 * channel.idx + 10 * (random.random() - 0.5)

    def PreReadAll(self):
        self.read_channels = {}

    def PreReadOne(self, ind):
        channel = self.channels[ind - 1]
        self.read_channels[ind] = channel

    def ReadAll(self):
        for channel in list(self.read_channels.values()):
            self._setChannelValue(channel)

    def ReadOne(self, ind):
        v = self.read_channels[ind].value
        return v



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer meta recorder
information"""

__all__ = ["RECORDER_TEMPLATE", "RecorderLibrary", "RecorderClass"]

__docformat__ = 'restructuredtext'

from sardana import InvalidId, ElementType
from sardana.sardanameta import SardanaLibrary, SardanaClass

#: String containing template code for a controller class
RECORDER_TEMPLATE = """class @recorder_name@(BaseFileRecorder):
    \"\"\"@recorder_name@ description.\"\"\"

"""


class RecorderLibrary(SardanaLibrary):
    """Object representing a python module containing recorder classes and/or
    recorder functions. Public members:

        - module - reference to python module
        - file_path - complete (absolute) path (with file name at the end)
        - file_name - file name (including file extension)
        - path - complete (absolute) path
        - name - (=module name) module name (without file extension)
        - recorder_list - list<RecorderClass>
        - exc_info - exception information if an error occurred when loading
                    the module"""

    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('macro_server')
        kwargs['elem_type'] = ElementType.RecorderLibrary
        SardanaLibrary.__init__(self, **kwargs)

    def serialize(self, *args, **kwargs):
        kwargs = SardanaLibrary.serialize(self, *args, **kwargs)
        kwargs['macro_server'] = self.get_manager().name
        kwargs['id'] = InvalidId
        return kwargs

    add_recorder = SardanaLibrary.add_meta_class
    get_recorder = SardanaLibrary.get_meta_class
    get_recorders = SardanaLibrary.get_meta_classes
    has_recorder = SardanaLibrary.has_meta_class

    @property
    def recorders(self):
        return self.meta_classes


class RecorderClass(SardanaClass):
    """Object representing a python recorder class.
       Public members:

           - name - class name
           - klass - python class object
           - lib - RecorderLibrary object representing the module where the
             recorder is."""

    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('macro_server')
        kwargs['elem_type'] = ElementType.RecorderClass
        SardanaClass.__init__(self, **kwargs)

    def serialize(self, *args, **kwargs):
        kwargs = SardanaClass.serialize(self, *args, **kwargs)
        kwargs['id'] = InvalidId
        kwargs['hints'] = self.code_object.hints
        kwargs['macro_server'] = self.get_manager().name
        return kwargs

    @property
    def recorder_class(self):
        return self.klass



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for environment macros"""

import unittest
from sardana.macroserver.macros.test import RunMacroTestCase, testRun


@testRun
class DumpenvTest(RunMacroTestCase, unittest.TestCase):
    """Test case for dumpenv macro
    """
    macro_name = "dumpenv"


@testRun
class LsvoTest(RunMacroTestCase, unittest.TestCase):
    """Test case for lsvo macro
    """
    macro_name = "lsvo"


@testRun(macro_params=["PosFormat", "3"])
class SetvoTest(RunMacroTestCase, unittest.TestCase):
    """Test case for setvo macro
    """
    macro_name = "setvo"


@testRun(macro_params=["PosFormat"])
class UsetvoTest(RunMacroTestCase, unittest.TestCase):
    """Test case for usetvo macro
    """
    macro_name = "usetvo"


@testRun
@testRun(macro_params=["ascan"])
@testRun(macro_params=["ascan", "dscan"])
class LsenvTest(RunMacroTestCase, unittest.TestCase):
    """Test case for lsvo macro
    """
    macro_name = "lsenv"


@testRun(macro_params=["MyEnvVar", "test.dat"])
class SenvTest(RunMacroTestCase, unittest.TestCase):
    """Test case for senv macro
    """
    macro_name = "senv"


@testRun(macro_params=["MyEnvVar"])
class UsenvTest(RunMacroTestCase, unittest.TestCase):
    """Test case for usenv macro
    """
    macro_name = "usenv"



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of the macroserver data type manager"""

__all__ = ["TypeManager"]

__docformat__ = 'restructuredtext'

import os
import inspect

from sardana.sardanamodulemanager import ModuleManager

from sardana.macroserver.msparameter import Type, ParamType, AbstractParamTypes
from sardana.macroserver.msmanager import MacroServerManager

_BASE_DIR = os.path.dirname(os.path.abspath(__file__))


class TypeManager(MacroServerManager):

    DEFAULT_TYPE_DIR = _BASE_DIR
    DEFAULT_TYPE_MODULES = 'basetypes',

    def reInit(self):
        if self.is_initialized():
            return

        # dict<str, dict<str,class<ParamType>>
        # key   - module name (without path and without extension)
        # value - a dict where:
        #         key   - type name
        #         value - class object implementing the Type
        self._modules = {}

        # dict<str, ParamType>
        # key   - type name
        # value - object which inherits from ParamType
        self._inst_dict = {}

        path = [self.DEFAULT_TYPE_DIR]
        for type_module in self.DEFAULT_TYPE_MODULES:
            self.reloadTypeModule(type_module, path=path)

        MacroServerManager.reInit(self)

    def cleanUp(self):
        if self.is_cleaned():
            return

        if self._modules:
            for _, types_dict in list(self._modules.items()):
                for type_name in types_dict:
                    Type.removeType(type_name)

        self._modules = None
        self._inst_dict = None

        MacroServerManager.cleanUp(self)

    def getTypeListObj(self):
        return self._type_list_obj

    def reloadTypeModule(self, module_name, path=None):
        """Loads/reloads the given module name"""
        #path = path or [ os.path.dirname(__file__) ]
        m = None
        try:
            m = ModuleManager().reloadModule(module_name, path)
        except:
            pass

        if m is None:
            if module_name in self._modules:
                self._modules.pop(module_name)
            return

        self._modules[module_name] = {}

        abs_file = inspect.getabsfile(m)
        ms = self.macro_server
        for name in dir(m):
            if name.startswith("_"):
                continue
            klass = getattr(m, name)
            try:
                if not issubclass(klass, ParamType):
                    continue
            except:
                continue
            if klass in AbstractParamTypes:
                continue
            if inspect.getabsfile(klass) != abs_file:
                continue

            t = klass(ms, name)
            self.addType(t)

    def addType(self, type_obj):
        type_name = type_obj.getName()
        type_class = type_obj.__class__
        module_name = type_obj.__module__

        mod_types = self._modules[module_name]

        # action = (((type_name in mod_types) and "Updating") \
        #          or "Adding")
        action = "Updating"
        self.debug("%s type %s", action, type_name)
        mod_types[type_name] = type_class
        self._inst_dict[type_name] = type_obj

        Type.addType(type_name)

    def getTypeListStr(self):
        type_list_basic, type_list_obj = [], []

        for _, type_class_dict in list(self._modules.items()):
            for tname, tklass in list(type_class_dict.items()):
                if tklass.hasCapability(ParamType.ItemList):
                    type_list_obj.append("%s*" % tname)
                else:
                    type_list_basic.append(tname)
        type_list = sorted(type_list_basic) + sorted(type_list_obj)

        return type_list

    def getTypeClass(self, type_name):
        for _, type_class_dict in list(self._modules.items()):
            tklass = type_class_dict.get(type_name)
            if tklass is None:
                continue
            return tklass
        return None

    def getTypeObj(self, type_name):
        return self._inst_dict.get(type_name)

    def getTypes(self):
        return self._inst_dict

    def getTypeNames(self):
        return list(self._inst_dict.keys())



#!/usr/bin/env python
# Copyright (c) 2008 Qtrac Ltd. All rights reserved.
# This program or module is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 2 of the License, or
# version 3 of the License, or (at your option) any later version. It is
# provided for educational purposes and is distributed in the hope that
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
# the GNU General Public License for more details.

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future_builtins import *
import sys
import os
import bisect
import codecs
import taurus.qt.qtgui.resource
import PyQt4.QtGui
from PyQt4.QtCore import (QAbstractItemModel, QModelIndex, QString,
                          QVariant, Qt, SIGNAL)

KEY, NODE = range(2)


class BranchNode(object):

    def __init__(self, name, parent=None):
        super(BranchNode, self).__init__()
        self.name = name
        self.parent = parent
        self.children = []

    def orderKey(self):
        return self.name.lower()

    def toString(self):
        return self.name

    def __len__(self):
        return len(self.children)

    def childAtRow(self, row):
        assert 0 <= row < len(self.children)
        return self.children[row][NODE]

    def rowOfChild(self, child):
        for i, item in enumerate(self.children):
            if item[NODE] == child:
                return i
        return -1

#     with sorting
#    def childWithKey(self, key):
#        if not self.children:
#            return None
#        # Causes a -3 deprecation warning. Solution will be to
#        # reimplement bisect_left and provide a key function.
#        i = bisect.bisect_left(self.children, (key, None))
#        if i < 0 or i >= len(self.children):
#            return None
#        if self.children[i][KEY] == key:
#            return self.children[i][NODE]
#        return None

    # without sorting
    def childWithKey(self, key):
        if not self.children:
            return None
        for i in range(len(self.children)):
            if self.children[i][KEY] == key:
                return self.children[i][NODE]
        return None

    def insertChild(self, child):
        child.parent = self
        # sorted
        #bisect.insort(self.children, (child.orderKey(), child))
        # unsorted
        self.children.append((child.orderKey(), child))

    def hasLeaves(self):
        if not self.children:
            return False
        return isinstance(self.children[0], LeafNode)


class LeafNode(object):

    def __init__(self, name, data, parent=None):
        super(LeafNode, self).__init__()
        self.parent = parent
        self.name = name
        self.data = data

    def orderKey(self):
        return "\t" + self.name.lower()

    def getName(self):
        return self.name

    def asRecord(self):
        return self.getName()

    def asData(self):
        return self.data


class SimpleTreeModel(QAbstractItemModel):

    def __init__(self, parent=None):
        super(SimpleTreeModel, self).__init__(parent)
        self.columns = 1
        self.root = BranchNode("")
        self.headers = ["Type"]

    def setHeaders(self, headers):
        self.headers = headers

    def getHeaders(self):
        return self.headers

    def addNodes(self, fields, callReset=True):
        root = self.root
        branch = None
        for i in range(len(fields)):
            key = fields[i].lower()
            branch = root.childWithKey(key)
            if branch is not None:
                root = branch
            else:
                branch = BranchNode(fields[i])
                root.insertChild(branch)
                root = branch
        if callReset:
            self.reset()
        self.headers = ["Type"]

    def addRecord(self, fields, data, callReset=True):
        root = self.root
        branch = None
        for i in range(len(fields) - 1):
            key = fields[i].lower()
            branch = root.childWithKey(key)
            if branch is not None:
                root = branch
            else:
                branch = BranchNode(fields[i])
                root.insertChild(branch)
                root = branch
        assert branch is not None
        item = fields[len(fields) - 1]
        branch.insertChild(LeafNode(item, data, branch))
        if callReset:
            self.reset()
        self.headers = ["Type"]

    def asRecord(self, index):
        leaf = self.nodeFromIndex(index)
        if leaf is not None and isinstance(leaf, LeafNode):
            return leaf.asRecord()
        return "None"

    def asData(self, index):
        leaf = self.nodeFromIndex(index)
        if leaf is not None and isinstance(leaf, LeafNode):
            return leaf.asData()
        else:
            return None

    def rowCount(self, parent):
        node = self.nodeFromIndex(parent)
        if node is None or isinstance(node, LeafNode):
            return 0
        return len(node)

    def columnCount(self, parent):
        return self.columns

    def data(self, index, role):
        if role == Qt.TextAlignmentRole:
            return QVariant(int(Qt.AlignTop | Qt.AlignLeft))

        if role == Qt.DecorationRole:
            node = self.nodeFromIndex(index)
            assert node is not None
            if isinstance(node, LeafNode):
                return QVariant(node.asData().get_icon())
            else:
                return QVariant()

        if role != Qt.DisplayRole:
            return QVariant()

        node = self.nodeFromIndex(index)
        assert node is not None

        if isinstance(node, BranchNode):
            if index.column() == 0:
                return QVariant(node.toString())
            else:
                return QVariant(QString(""))

        if isinstance(node, LeafNode):
            return QVariant(QString(node.getName()))

    def headerData(self, section, orientation, role):
        if (orientation == Qt.Horizontal and
                role == Qt.DisplayRole):
            #assert 0 <= section <= len(self.headers)
            return QVariant(self.headers[section])
        return QVariant()

    def index(self, row, column, parent):
        assert self.root
        branch = self.nodeFromIndex(parent)
        assert branch is not None
        return self.createIndex(row, column,
                                branch.childAtRow(row))

    def parent(self, child):
        node = self.nodeFromIndex(child)
        if node is None:
            return QModelIndex()
        parent = node.parent
        if parent is None:
            return QModelIndex()
        grandparent = parent.parent
        if grandparent is None:
            return QModelIndex()
        row = grandparent.rowOfChild(parent)
        assert row != -1
        return self.createIndex(row, 0, parent)

    def nodeFromIndex(self, index):
        if index.isValid():
            return index.internalPointer()
        else:
            return self.root

        # return (index.internalPointer()
        #        if index.isValid() else self.root)######################



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module provides widget for configuring the data acquisition and display of an experiment"""

__all__ = ["ExpDescriptionEditor"]


import json
from taurus.external.qt import Qt, QtCore, QtGui, compat
import copy
import taurus
import taurus.core
from taurus.qt.qtgui.base import TaurusBaseWidget

from sardana.taurus.qt.qtcore.tango.sardana.model import SardanaBaseProxyModel, SardanaTypeTreeItem
from sardana.sardanadefs import ElementType, TYPE_ACQUIRABLE_ELEMENTS
from taurus.qt.qtgui.util.ui import UILoadable

# Using a plain model and filtering and checking
# 'Acquirable' in item.itemData().interfaces is more elegant,
# but things don't get properly sorted...

# from taurus.qt.qtcore.tango.sardana.model import SardanaElementPlainModel


def _to_fqdn(name, logger=None):
    """Helper to convert name into a FQDN URI reference. Works for devices
    and attributes.
    Prior to sardana 2.4.0 Pool element references were using not unique full
    names e.g. pc255:10000/motor/motctrl20/1. This helper converts them into
    FQDN URI references e.g. tango://pc255.cells.es:10000/motor/motctrl20/1.
    It is similarly solved for attribute names.
    """
    full_name = name
    # try to use Taurus 4 to retrieve FQDN URI name
    try:
        from taurus.core.tango.tangovalidator import TangoDeviceNameValidator
        try:
            full_name, _, _ = TangoDeviceNameValidator().getNames(name)
        # it is not a device try as an attribute
        except Exception:
            from taurus.core.tango.tangovalidator import \
                TangoAttributeNameValidator
            full_name, _, _ = TangoAttributeNameValidator().getNames(name)
    # if Taurus3 in use just continue
    except ImportError:
        pass
    if full_name != name and logger:
        msg = ("PQDN full name is deprecated in favor of FQDN full name. "
               "Re-apply pre-scan snapshot configuration in order to "
               "upgrade.")
        logger.warning(msg)
    return full_name


class SardanaAcquirableProxyModel(SardanaBaseProxyModel):
    #    ALLOWED_TYPES = 'Acquirable'
    #
    #    def filterAcceptsRow(self, sourceRow, sourceParent):
    #        sourceModel = self.sourceModel()
    #        idx = sourceModel.index(sourceRow, 0, sourceParent)
    #        item = idx.internalPointer()
    #        return 'Acquirable' in item.itemData().interfaces

    #    ALLOWED_TYPES = ['Motor', 'CTExpChannel', 'ZeroDExpChannel', 'OneDExpChannel',
    #                     'TwoDExpChannel', 'ComChannel', 'IORegister', 'PseudoMotor',
    #                     'PseudoCounter']

    ALLOWED_TYPES = [ElementType[t] for t in TYPE_ACQUIRABLE_ELEMENTS]

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()
        if isinstance(treeItem, SardanaTypeTreeItem):
            return treeItem.itemData() in self.ALLOWED_TYPES
        return True


def find_diff(first, second):
    """
    Return a dict of keys that differ with another config object.  If a value
    is not found in one fo the configs, it will be represented by KEYNOTFOUND.
    :param first: Fist configuration to diff.
    :param second: Second configuration to diff.
    :return: Dict of Key => (first.val, second.val)
    """

    KEYNOTFOUNDIN1 = 'KeyNotFoundInRemote'
    KEYNOTFOUNDIN2 = 'KeyNotFoundInLocal'

    # The GUI can not change these keys. They are changed by the server.
    SKIPKEYS = ['_controller_name', 'description', 'timer', 'monitor', 'ndim',
                'source']

    # These keys can have a list as value.
    SKIPLIST = ['scanfile', 'plot_axes', 'prescansnapshot', 'shape']

    DICT_TYPES = [taurus.core.util.containers.CaselessDict, dict]
    diff = {}
    sd1 = set(first)
    sd2 = set(second)

    # Keys missing in the second dict
    for key in sd1.difference(sd2):
        if key in SKIPKEYS:
            continue
        diff[key] = (first[key], KEYNOTFOUNDIN2)
    # Keys missing in the first dict
    for key in sd2.difference(sd1):
        if key in SKIPKEYS:
            continue
        diff[key] = (KEYNOTFOUNDIN1, second[key])

    # Check for differences
    for key in sd1.intersection(sd2):
        if key in SKIPKEYS:
            continue
        value1 = first[key]
        value2 = second[key]
        if type(value1) in DICT_TYPES:
            try:
                idiff = find_diff(value1, value2)
            except Exception:
                idiff = 'Error on processing'
            if len(idiff) > 0:
                diff[key] = idiff
        elif isinstance(value1, list) and key.lower() not in SKIPLIST:
            ldiff = []
            for v1, v2 in zip(value1, value2):
                try:
                    idiff = find_diff(v1, v2)
                except Exception:
                    idiff = 'Error on processing'
                ldiff.append(idiff)
            if len(ldiff) > 0:
                diff[key] = ldiff
        else:
            if value1 != value2:
                diff[key] = (first[key], second[key])
    return diff


@UILoadable(with_ui='ui')
class ExpDescriptionEditor(Qt.QWidget, TaurusBaseWidget):
    '''
    A widget for editing the configuration of a experiment (measurement groups,
    plot and storage parameters, etc).

    It receives a Sardana Door name as its model and gets/sets the configuration
    using the `ExperimentConfiguration` environmental variable for that Door.
    '''

    createExpConfChangedDialog = Qt.pyqtSignal()
    experimentConfigurationChanged = Qt.pyqtSignal(compat.PY_OBJECT)

    def __init__(self, parent=None, door=None, autoUpdate=False):
        Qt.QWidget.__init__(self, parent)
        TaurusBaseWidget.__init__(self, 'ExpDescriptionEditor')
        self.loadUi()
        self.ui.buttonBox.setStandardButtons(
            Qt.QDialogButtonBox.Reset | Qt.QDialogButtonBox.Apply)
        self.ui.buttonBox.button(Qt.QDialogButtonBox.Reset).setText('Reload')

        newperspectivesDict = copy.deepcopy(
            self.ui.sardanaElementTree.KnownPerspectives)
        #newperspectivesDict[self.ui.sardanaElementTree.DftPerspective]['model'] = [SardanaAcquirableProxyModel, SardanaElementPlainModel]
        newperspectivesDict[self.ui.sardanaElementTree.DftPerspective][
            'model'][0] = SardanaAcquirableProxyModel
        # assign a copy because if just a key of this class memberwas modified,
        # all instances of this class would be affected
        self.ui.sardanaElementTree.KnownPerspectives = newperspectivesDict
        self.ui.sardanaElementTree._setPerspective(
            self.ui.sardanaElementTree.DftPerspective)

        self._localConfig = None
        self._originalConfiguration = None
        self._dirty = False
        self._dirtyMntGrps = set()

        self._autoUpdate = False
        self._warningWidget = None
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)
        self._autoUpdateAction = Qt.QAction("Auto update", self)
        self._autoUpdateAction.setCheckable(True)
        self._autoUpdateAction.toggled.connect(self.setAutoUpdate)
        self.addAction(self._autoUpdateAction)
        self._autoUpdateAction.setChecked(autoUpdate)
        self.registerConfigProperty(
            self._autoUpdateAction.isChecked,
            self._autoUpdateAction.setChecked,
            "autoUpdate")

        # Pending event variables
        self._expConfChangedDialog = None

        self.createExpConfChangedDialog.connect(
            self._createExpConfChangedDialog)
        self.ui.activeMntGrpCB.activated['QString'].connect(
            self.changeActiveMntGrp)
        self.ui.createMntGrpBT.clicked.connect(
            self.createMntGrp)
        self.ui.deleteMntGrpBT.clicked.connect(
            self.deleteMntGrp)
        self.ui.compressionCB.currentIndexChanged['int'].connect(
            self.onCompressionCBChanged)
        self.ui.pathLE.textEdited.connect(
            self.onPathLEEdited)
        self.ui.filenameLE.textEdited.connect(
            self.onFilenameLEEdited)
        self.ui.channelEditor.getQModel().dataChanged.connect(
            self._updateButtonBox)
        self.ui.channelEditor.getQModel().modelReset.connect(
            self._updateButtonBox)
        preScanList = self.ui.preScanList
        preScanList.dataChangedSignal.connect(self.onPreScanSnapshotChanged)
        self.ui.choosePathBT.clicked.connect(
            self.onChooseScanDirButtonClicked)

        if door is not None:
            self.setModel(door)

        self.ui.buttonBox.clicked.connect(self.onDialogButtonClicked)

        # Taurus Configuration properties and delegates
        self.registerConfigDelegate(self.ui.channelEditor)

    def setAutoUpdate(self, auto_update):
        if auto_update and not self._autoUpdate:
            self._warningWidget = self._getWarningWidget()
            self.ui.verticalLayout_3.insertWidget(0, self._warningWidget)
        if not auto_update and self._autoUpdate:
            self.ui.verticalLayout_3.removeWidget(self._warningWidget)
            self._warningWidget.deleteLater()
            self._warningWidget = None
        self._autoUpdate = auto_update

    def _getWarningWidget(self):
        w = Qt.QWidget()
        layout = QtGui.QHBoxLayout()
        w.setLayout(layout)
        icon = QtGui.QIcon.fromTheme('dialog-warning')
        pixmap = QtGui.QPixmap(icon.pixmap(QtCore.QSize(32, 32)))
        label_icon = QtGui.QLabel()
        label_icon.setPixmap(pixmap)
        label = QtGui.QLabel('This experiment configuration dialog '
                             'updates automatically on external changes!')
        layout.addWidget(label_icon)
        layout.addWidget(label)
        layout.addStretch(1)
        return w

    def _getResumeText(self):
        msg_resume = '<p> Summary of differences: <ul>'
        mnt_grps = ''
        envs = ''
        for key in self._diff:
            if key == 'MntGrpConfigs':
                for names in self._diff['MntGrpConfigs']:
                    if mnt_grps != '':
                        mnt_grps += ', '
                    mnt_grps += '<b>{0}</b>'.format(names)
            else:
                if envs != '':
                    envs += ', '
                envs += '<b>{0}</b>'.format(key)
        values = ''
        if mnt_grps != '':
            values += '<li> Measurement Groups: {0}</li>'.format(mnt_grps)
        if envs != '':
            values += '<li> Enviroment variables: {0}</li>'.format(envs)

        msg_resume += values
        msg_resume += ' </ul> </p>'
        return msg_resume

    def _getDetialsText(self):
        msg_detials = 'Changes {key: [external, local], ...}\n'
        msg_detials += json.dumps(self._diff, sort_keys=True)
        return msg_detials

    def _createExpConfChangedDialog(self):
        msg_details = self._getDetialsText()
        msg_info = self._getResumeText()
        self._expConfChangedDialog = Qt.QMessageBox()
        self._expConfChangedDialog.setIcon(Qt.QMessageBox.Warning)
        self._expConfChangedDialog.setWindowTitle('External Changes')
        # text = '''
        # <p align='justify'>
        # The experiment configuration has been modified externally.<br/>
        # You can either:<br/> <l1><b>Load</b> the new configuration from the
        # door
        # (discarding local changes) or <b>Keep</b> your local configuration
        # (would eventually overwrite the external changes when applying).
        # </p>'''
        text = '''
        <p>The experiment configuration has been modified externally.<br>
        You can either:
        <ul>
        <li><strong>Load </strong>the new external configuration</li>
        <li><strong>Keep </strong>your local expconf configuration<br>
        (It can be eventually applied)</li>
        </ul></p>
        '''
        self._expConfChangedDialog.setText(text)
        self._expConfChangedDialog.setTextFormat(QtCore.Qt.RichText)
        self._expConfChangedDialog.setInformativeText(msg_info)
        self._expConfChangedDialog.setDetailedText(msg_details)
        self._expConfChangedDialog.setStandardButtons(Qt.QMessageBox.Ok |
                                                      Qt.QMessageBox.Cancel)
        btn_ok = self._expConfChangedDialog.button(Qt.QMessageBox.Ok)
        btn_ok.setText('Load')
        btn_cancel = self._expConfChangedDialog.button(Qt.QMessageBox.Cancel)
        btn_cancel.setText('Keep')
        result = self._expConfChangedDialog.exec_()
        self._expConfChangedDialog = None
        if result == Qt.QMessageBox.Ok:
            self._reloadConf(force=True)
        elif result == Qt.QMessageBox.Cancel:
            self.ui.buttonBox.setEnabled(True)

    @QtCore.pyqtSlot()
    def _experimentConfigurationChanged(self):
        self._diff = ''
        try:
            self._diff = self._getDiff()
        except Exception as e:
            raise RuntimeError('Error on processing! {0}'.format(e))

        if len(self._diff) > 0:
            if self._autoUpdate:
                self._reloadConf(force=True)
            else:
                if self._expConfChangedDialog is None:
                    if hasattr(self, 'createExpConfChangedDialog'):
                        self.createExpConfChangedDialog.emit()
                else:
                    msg_details = self._getDetialsText()
                    msg_info = self._getResumeText()
                    self._expConfChangedDialog.setInformativeText(msg_info)
                    self._expConfChangedDialog.setDetailedText(msg_details)

    def _getDiff(self):
        door = self.getModelObj()
        if door is None:
            return []

        new_conf = door.getExperimentConfiguration()
        old_conf = self._localConfig
        return find_diff(new_conf, old_conf)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return taurus.core.taurusdevice.TaurusDevice

    def onChooseScanDirButtonClicked(self):
        ret = Qt.QFileDialog.getExistingDirectory(
            self, 'Choose directory for saving files', self.ui.pathLE.text())
        if ret:
            self.ui.pathLE.setText(ret)
            self.ui.pathLE.textEdited.emit(ret)

    def onDialogButtonClicked(self, button):
        role = self.ui.buttonBox.buttonRole(button)
        if role == Qt.QDialogButtonBox.ApplyRole:
            if not self.writeExperimentConfiguration(ask=False):
                self._reloadConf(force=True)
        elif role == Qt.QDialogButtonBox.ResetRole:
            self._reloadConf()

    def closeEvent(self, event):
        '''This event handler receives widget close events'''
        if self.isDataChanged():
            self.writeExperimentConfiguration(ask=True)
        Qt.QWidget.closeEvent(self, event)

    def setModel(self, model):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        TaurusBaseWidget.setModel(self, model)
        self._reloadConf(force=True)
        # set the model of some child widgets
        door = self.getModelObj()
        if door is None:
            return
        # @todo: get the tghost from the door model instead
        tghost = taurus.Authority().getNormalName()
        msname = door.macro_server.getFullName()
        self.ui.taurusModelTree.setModel(tghost)
        self.ui.sardanaElementTree.setModel(msname)
        door.experimentConfigurationChanged.connect(
            self._experimentConfigurationChanged)

    def _reloadConf(self, force=False):
        if not force and self.isDataChanged():
            op = Qt.QMessageBox.question(self, "Reload info from door",
                                         "If you reload, all current experiment configuration changes will be lost. Reload?",
                                         Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
            if op != Qt.QMessageBox.Yes:
                return
        door = self.getModelObj()
        if door is None:
            return
        conf = door.getExperimentConfiguration()
        self._originalConfiguration = copy.deepcopy(conf)
        self.setLocalConfig(conf)
        self._setDirty(False)
        self._dirtyMntGrps = set()
        # set a list of available channels
        avail_channels = {}
        for ch_info in \
                list(door.macro_server.getExpChannelElements().values()):
            avail_channels[ch_info.full_name] = ch_info.getData()
        self.ui.channelEditor.getQModel().setAvailableChannels(avail_channels)
        # set a list of available triggers
        avail_triggers = {'software': {"name": "software"}}
        tg_elements = door.macro_server.getElementsOfType('TriggerGate')
        for tg_info in list(tg_elements.values()):
            avail_triggers[tg_info.full_name] = tg_info.getData()
        self.ui.channelEditor.getQModel().setAvailableTriggers(avail_triggers)
        self.experimentConfigurationChanged.emit(copy.deepcopy(conf))

    def _setDirty(self, dirty):
        self._dirty = dirty
        self._updateButtonBox()

    def isDataChanged(self):
        """Tells if the local data has been modified since it was last refreshed

        :return: (bool) True if he local data has been modified since it was last refreshed
        """
        return bool(self._dirty or self.ui.channelEditor.getQModel().isDataChanged() or self._dirtyMntGrps)

    def _updateButtonBox(self, *args, **kwargs):
        self.ui.buttonBox.setEnabled(self.isDataChanged())

    def getLocalConfig(self):
        return self._localConfig

    def setLocalConfig(self, conf):
        '''gets a ExpDescription dictionary and sets up the widget'''

        self._localConfig = conf

        # set the Channel Editor
        activeMntGrpName = self._localConfig['ActiveMntGrp'] or ''
        if activeMntGrpName in self._localConfig['MntGrpConfigs']:
            mgconfig = self._localConfig['MntGrpConfigs'][activeMntGrpName]
            self.ui.channelEditor.getQModel().setDataSource(mgconfig)
        else:
            self.ui.channelEditor.getQModel().setDataSource({})

        # set the measurement group ComboBox
        self.ui.activeMntGrpCB.clear()
        mntGrpLabels = []
        for _, mntGrpConf in list(self._localConfig['MntGrpConfigs'].items()):
            # get labels to visualize names with lower and upper case
            mntGrpLabels.append(mntGrpConf['label'])
        self.ui.activeMntGrpCB.addItems(sorted(mntGrpLabels))
        idx = self.ui.activeMntGrpCB.findText(activeMntGrpName,
                                              # case insensitive find
                                              Qt.Qt.MatchFixedString)
        self.ui.activeMntGrpCB.setCurrentIndex(idx)

        # set the system snapshot list
        # I get it before clearing because clear() changes the _localConfig
        psl = self._localConfig.get('PreScanSnapshot')
        # TODO: For Taurus 4 compatibility
        psl_fullname = []
        for name, display in psl:
            name = _to_fqdn(name, self)
            psl_fullname.append((name, display))

        self.ui.preScanList.clear()
        self.ui.preScanList.addModels(psl_fullname)

        # other settings
        self.ui.filenameLE.setText(", ".join(self._localConfig['ScanFile']))
        self.ui.pathLE.setText(self._localConfig['ScanDir'] or '')
        self.ui.compressionCB.setCurrentIndex(
            self._localConfig['DataCompressionRank'] + 1)

    def writeExperimentConfiguration(self, ask=True):
        '''sends the current local configuration to the door

        :param ask: (bool) If True (default) prompts the user before saving.
        '''

        if ask:
            op = Qt.QMessageBox.question(self, "Save configuration?",
                                         'Do you want to save the current configuration?\n(if not, any changes will be lost)',
                                         Qt.QMessageBox.Yes | Qt.QMessageBox.No)
            if op != Qt.QMessageBox.Yes:
                return False

        conf = self.getLocalConfig()

        # make sure that no empty measurement groups are written
        for mgname, mgconfig in list(conf.get('MntGrpConfigs', {}).items()):
            if mgconfig is not None and not mgconfig.get('controllers'):
                mglabel = mgconfig['label']
                Qt.QMessageBox.information(self, "Empty Measurement group",
                                           "The measurement group '%s' is empty. Fill it (or delete it) before applying" % mglabel,
                                           Qt.QMessageBox.Ok)
                self.changeActiveMntGrp(mgname)
                return False

        # check if the currently displayed mntgrp is changed
        if self.ui.channelEditor.getQModel().isDataChanged():
            self._dirtyMntGrps.add(self._localConfig['ActiveMntGrp'])

        door = self.getModelObj()
        try:
            door.setExperimentConfiguration(conf, mnt_grps=self._dirtyMntGrps)
        except Exception as e:
            Qt.QMessageBox.critical(self, 'Wrong configuration',
                                    '{0}'.format(e))
            return False
        self._originalConfiguration = copy.deepcopy(conf)
        self._dirtyMntGrps = set()
        self.ui.channelEditor.getQModel().setDataChanged(False)
        self._setDirty(False)
        self.experimentConfigurationChanged.emit(copy.deepcopy(conf))
        return True

    @Qt.pyqtSlot('QString')
    def changeActiveMntGrp(self, activeMntGrpName):
        if self._localConfig is None:
            return
        if activeMntGrpName == self._localConfig['ActiveMntGrp']:
            return  # nothing changed
        if activeMntGrpName not in self._localConfig['MntGrpConfigs']:
            raise KeyError('Unknown measurement group "%s"' % activeMntGrpName)

        # add the previous measurement group to the list of "dirty" groups if
        # something was changed
        if self.ui.channelEditor.getQModel().isDataChanged():
            self._dirtyMntGrps.add(self._localConfig['ActiveMntGrp'])

        self._localConfig['ActiveMntGrp'] = activeMntGrpName

        i = self.ui.activeMntGrpCB.findText(activeMntGrpName,
                                            # case insensitive find
                                            Qt.Qt.MatchFixedString)
        self.ui.activeMntGrpCB.setCurrentIndex(i)
        mgconfig = self._localConfig['MntGrpConfigs'][activeMntGrpName]
        self.ui.channelEditor.getQModel().setDataSource(mgconfig)
        self._setDirty(True)

    def createMntGrp(self):
        '''creates a new Measurement Group'''

        if self._localConfig is None:
            return

        mntGrpName, ok = Qt.QInputDialog.getText(self, "New Measurement Group",
                                                 "Enter a name for the new measurement Group")
        if not ok:
            return
        mntGrpName = str(mntGrpName)

        # check that the given name is not an existing pool element
        ms = self.getModelObj().macro_server
        poolElementNames = [
            v.name for v in
            list(ms.getElementsWithInterface("PoolElement").values())]
        while mntGrpName in poolElementNames:
            msg = ("The name '%s' already is used for another pool element. "
                   "Please Choose a different one." % mntGrpName)
            Qt.QMessageBox.warning(self, "Cannot create Measurement group",
                                   msg, Qt.QMessageBox.Ok)
            msg = "Enter a name for the new measurement Group"
            mntGrpName, ok = Qt.QInputDialog.getText(self,
                                                     "New Measurement Group",
                                                     msg,
                                                     Qt.QLineEdit.Normal,
                                                     mntGrpName)
            if not ok:
                return
            mntGrpName = str(mntGrpName)

        # check that the measurement group is not already in the localConfig
        msg = ('A measurement group named "%s" already exists. A new one '
               'will not be created' % mntGrpName)
        if mntGrpName in self._localConfig['MntGrpConfigs']:
            Qt.QMessageBox.warning(self, "%s already exists" % mntGrpName,
                                   msg)
            return

        # add an empty configuration dictionary to the local config
        mgconfig = {'label': mntGrpName, 'controllers': {}}
        self._localConfig['MntGrpConfigs'][mntGrpName] = mgconfig
        # add the new measurement group to the list of "dirty" groups
        self._dirtyMntGrps.add(mntGrpName)
        # add the name to the combobox
        self.ui.activeMntGrpCB.addItem(mntGrpName)
        # make it the Active MntGrp
        self.changeActiveMntGrp(mntGrpName)

    def deleteMntGrp(self):
        '''creates a new Measurement Group'''
        activeMntGrpName = str(self.ui.activeMntGrpCB.currentText())
        op = Qt.QMessageBox.question(self, "Delete Measurement Group",
                                     "Remove the measurement group '%s'?" % activeMntGrpName,
                                     Qt.QMessageBox.Yes | Qt.QMessageBox.Cancel)
        if op != Qt.QMessageBox.Yes:
            return
        currentIndex = self.ui.activeMntGrpCB.currentIndex()
        if self._localConfig is None:
            return
        if activeMntGrpName not in self._localConfig['MntGrpConfigs']:
            raise KeyError('Unknown measurement group "%s"' % activeMntGrpName)

        # add the current measurement group to the list of "dirty" groups
        self._dirtyMntGrps.add(activeMntGrpName)

        self._localConfig['MntGrpConfigs'][activeMntGrpName] = None
        self.ui.activeMntGrpCB.setCurrentIndex(-1)
        self.ui.activeMntGrpCB.removeItem(currentIndex)
        self.ui.channelEditor.getQModel().setDataSource({})
        self._setDirty(True)

    @Qt.pyqtSlot('int')
    def onCompressionCBChanged(self, idx):
        if self._localConfig is None:
            return
        self._localConfig['DataCompressionRank'] = idx - 1
        self._setDirty(True)

    def onPathLEEdited(self, text):
        self._localConfig['ScanDir'] = str(text)
        self._setDirty(True)

    def onFilenameLEEdited(self, text):
        self._localConfig['ScanFile'] = [v.strip()
                                         for v in str(text).split(',')]
        self._setDirty(True)

    def onPreScanSnapshotChanged(self, items):
        door = self.getModelObj()
        ms = door.macro_server
        preScanList = []
        for e in items:
            nfo = ms.getElementInfo(e.src)
            if nfo is None:
                full_name = e.src
                display = e.display
            else:
                full_name = nfo.full_name
                display = nfo.name
            preScanList.append((full_name, display))
        self._localConfig['PreScanSnapshot'] = preScanList
        self._setDirty(True)


def demo(model=None, autoUpdate=False):
    """Experiment configuration"""
    #w = main_ChannelEditor()
    w = ExpDescriptionEditor(autoUpdate=autoUpdate)
    if model is None:
        from sardana.taurus.qt.qtgui.extra_macroexecutor import \
            TaurusMacroConfigurationDialog
        dialog = TaurusMacroConfigurationDialog(w)
        accept = dialog.exec_()
        if accept:
            model = str(dialog.doorComboBox.currentText())
    if model is not None:
        w.setModel(model)
    return w


def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None
    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <door name>"
        parser.add_option('--auto-update', dest='auto_update',
                          action='store_true',
                          help='Set auto update of experiment configuration')

        app = Application(app_name="Exp. Description demo", app_version="1.0",
                          org_domain="Sardana", org_name="Tango community",
                          cmd_line_parser=parser)

    args = app.get_command_line_args()
    opt = app.get_command_line_options()

    if len(args) == 1:
        auto_update = opt.auto_update is not None
        w = demo(model=args[0], autoUpdate=auto_update)
    else:
        w = demo()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w


if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for PoolTriggerGate"""

__all__ = ["PoolTriggerGate"]

__docformat__ = 'restructuredtext'

from sardana import ElementType

from sardana.pool.poolelement import PoolElement
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool.poolsynchronization import PoolSynchronization


class Index(SardanaAttribute):
    """ Index of the trigger/gate event being currently generated.

    .. note::
        The Index class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the module) may occur if
        deemed necessary by the core developers.
    """

    def __init__(self, *args, **kwargs):
        super(Index, self).__init__(*args, **kwargs)


class PoolTriggerGate(PoolElement):

    def __init__(self, **kwargs):
        kwargs['elem_type'] = ElementType.TriggerGate
        PoolElement.__init__(self, **kwargs)
        tggen_name = "%s.TGGeneration" % self._name
        self.set_action_cache(PoolSynchronization(self, name=tggen_name))
        self._index = Index(self)

    # --------------------------------------------------------------------------
    # index
    # --------------------------------------------------------------------------

    def get_index_attribute(self):
        """Returns the index attribute object for this trigger/gate

        :return: the index attribute
        :rtype: :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        return self._index

    # --------------------------------------------------------------------------
    # default acquisition channel
    # --------------------------------------------------------------------------

    def get_default_attribute(self):
        return self.get_index_attribute()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__docformat__ = 'restructuredtext'

import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.base import TaurusBaseWidget
import time
import PyTango

import taurus.core

from taurus.external.qt import QtCore, QtGui

from taurus.qt.qtgui.util.ui import UILoadable

from sardana.taurus.core.tango.sardana.macroserver import registerExtensions


class SignalComboBox(Qt.QComboBox, TaurusBaseWidget):
    """ComboBox representing list of possible signals"""

    def __init__(self, parent=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWidget, name='')
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContentsOnFirstShow)
        self.setToolTip("Choose a signal ...")
        QtCore.QMetaObject.connectSlotsByName(self)

    def loadSignals(self, signals):
        self.clear()
        self.addItems(signals)


@UILoadable(with_ui="_ui")
class SelectSignal(TaurusWidget):

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent=None, designMode=designMode)

        self.loadUi(filename="selectsignal.ui")

        self.signalComboBox = SignalComboBox(self)
        self.signalComboBox.setGeometry(QtCore.QRect(70, 50, 161, 27))
        self.signalComboBox.setObjectName("SignalcomboBox")

        self.signalComboBox.currentIndexChanged['QString'].connect(
            self.onSignalChanged)

        self.doorName = None
        self.door_device = None

        registerExtensions()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'selectsignal'
        ret['group'] = 'Taurus Containers'
        ret['container'] = ':/designer/frame.png'
        ret['container'] = False
        return ret

    def update_signals(self, doorname=''):
        if self.doorName != doorname:
            self.doorName = doorname
            self.door_device = taurus.Device(self.doorName)
            print("Create door_device with name " + str(self.doorName))

        if self.doorName is not None:
            signals = []
            conf = self.door_device.getExperimentConfiguration()
            mg_name = conf['ActiveMntGrp']
            mg = taurus.Device(mg_name)
            signals = mg.ElementList
            self.signalComboBox.loadSignals(signals)

    def onSignalChanged(self, signalname):
        self._ui.SignallineEdit.setText(signalname)


def main():
    app = Qt.QApplication(sys.argv)
    w = SelectSignal()
    w.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



import sys
import os

from PyQt4 import QtGui, QtCore, Qt

import taurus
import wiz

from taurus.core.utils import Enumeration


class SelectSardanaBasePage(wiz.SardanaBasePage):

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)

        self['sardana'] = self._getSardana

        self.setSubTitle('Please select the Sardana instance')

        panel = self.getPanelWidget()

        layout = QtGui.QFormLayout()

        self.sardanaCB = QtGui.QComboBox(panel)
        self.sardanaCB.setDuplicatesEnabled(False)
        self.sardanaCB.setEditable(False)

        layout.addRow("&Sardana instance", self.sardanaCB)

        panel.setLayout(layout)

        self.connect(self.sardanaCB,
                     QtCore.SIGNAL('currentIndexChanged(int)'),
                     QtCore.SIGNAL('completeChanged()'))

    def _getSardana(self):
        sardana = str(self.sardanaCB.currentText())
        return sardana

    def sardana(self):
        return self.wizard()['sardana']

    def db(self):
        return self.wizard()['db']

    def isComplete(self):
        idx = self.sardanaCB.currentIndex()
        if idx >= 0:
            self.setStatus('')
            return True
        self.setStatus('No instance selected')
        return False

    def _fillCB(self):
        self.sardanaCB.clear()
        db = self.db()
        services = db.get_service_list('Sardana.*')
        service_instances = []
        for service in services:
            service_instances.append(service.split('/', 1)[1])
        self.sardanaCB.addItems(service_instances)

    def initializePage(self):
        wiz.SardanaBasePage.initializePage(self)
        self._fillCB()


def t1(tg_host=None):

    SardanaExamplePages = Enumeration(
        'SardanaExamplePages', ('IntroPage', 'TangoPage', 'SardanaPage'))

    class SardanaExampleIntroPage(wiz.SardanaIntroBasePage):

        def setNextPageId(self, id):
            self._nextPageId = id

        def nextId(self):
            return self._nextPageId

    from tango_host_page import SelectTangoHostBasePage

    class SelectTangoHostExamplePage(SelectTangoHostBasePage):

        def setNextPageId(self, id):
            self._nextPageId = id

        def nextId(self):
            return self._nextPageId

    app = QtGui.QApplication([])
    QtCore.QResource.registerResource(wiz.get_resources())

    w = wiz.SardanaBaseWizard()

    intro = SardanaExampleIntroPage()
    w.setPage(SardanaExamplePages.IntroPage, intro)

    curr_page = intro
    if tg_host is None:
        curr_page.setNextPageId(SardanaExamplePages.TangoPage)
        from tango_host_page import SelectTangoHostBasePage
        tg_host_page = SelectTangoHostBasePage()
        w.setPage(SardanaExamplePages.TangoPage, tg_host_page)
        curr_page = tg_host_page
    else:
        w['db'] = lambda: taurus.Database(tg_host)

    curr_page.setNextPageId(SardanaExamplePages.SardanaPage)
    sardana_page = SelectSardanaBasePage()
    w.setPage(SardanaExamplePages.SardanaPage, sardana_page)

    w.show()
    sys.exit(app.exec_())


def main():
    import taurus
    taurus.setLogLevel(taurus.Warning)
    tg_host = None
    if len(sys.argv) > 1:
        tg_host = sys.argv[1]
    t1(tg_host=tg_host)

if __name__ == '__main__':
    main()



from sardana.macroserver.macro import Type, Macro

FAIL_MSG = "Parsing or decoding failed (result: %r; expected: %r)"


class runMacro(Macro):
    """
    Macro to test the parameters parsing/decoding using the macro API,
    'runMacro'
    """
    def run(self, *args):
        params = expected_params = (99, [1., 2.])
        macro, _ = self.prepareMacro("pt6_base", *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = (99, 1., 2.)
        macro, _ = self.prepareMacro("pt6_base", *params)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 [1. 2.]"
        macro, _ = self.prepareMacro(params)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 1. 2."
        macro, _ = self.prepareMacro(params)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = expected_params = ([92], True)
        macro, _ = self.prepareMacro("pt10_base", *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = (91, True)
        expected_params = ([91], True)
        macro, _ = self.prepareMacro("pt10_base", *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base [91] True"
        expected_params = ([91], True)
        macro, _ = self.prepareMacro(params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base 91 True"
        expected_params = ([91], True)
        macro, _ = self.prepareMacro(params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg


class createMacro(Macro):
    """
    Macro to test the parameters parsing/decoding using the macro API,
    'createMacro'
    """
    def run(self, *args):

        params = expected_params = (99, [1., 2.])
        macro, pars = self.createMacro('pt6_base', *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = (99, 1., 2.)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        macro, pars = self.createMacro('pt6_base', *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 [1. 2.]"
        macro, _ = self.createMacro(params)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 1. 2."
        macro, _ = self.createMacro(params)
        expected_params = (99, [1., 2.])
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = expected_params = ([92], True)
        macro, _ = self.createMacro("pt10_base", *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = (91, True)
        expected_params = ([91.], True)
        macro, _ = self.createMacro("pt10_base", *params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base [91] True"
        expected_params = ([91], True)
        macro, _ = self.createMacro(params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base 91 True"
        expected_params = ([91], True)
        macro, _ = self.createMacro(params)
        self.runMacro(macro)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg


class execMacro(Macro):
    """
    Macro to test the parameters parsing/decoding using the macro API,
    'execMacro'
    """
    def run(self, *args):
        params = expected_params = (99, [1., 2.])
        macro = self.execMacro('pt6_base', *params)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = (99, 1., 2.)
        macro = self.execMacro('pt6_base', *params)
        result = macro.data
        expected_params = (99, [1., 2.])
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 [1 2]"
        macro = self.execMacro(params)
        expected_params = (99, [1., 2.])
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt6_base 99 1 2"
        macro = self.execMacro(params)
        expected_params = (99, [1., 2.])
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = expected_params = ([92], True)
        macro = self.execMacro('pt10_base', *params)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        expected_params = ([99.], True)
        macro = self.execMacro("pt10_base [99] True")
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        expected_params = ([999.], True)
        macro = self.execMacro("pt10_base", 999, True)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base [91] True"
        expected_params = ([91], True)
        macro = self.execMacro(params)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg

        params = "pt10_base 91 True"
        expected_params = ([91], True)
        macro = self.execMacro(params)
        result = macro.data
        msg = FAIL_MSG % (result, expected_params)
        assert expected_params == result, msg


class pt6_base(Macro):
    """Macro with a number parameter followed by a list of numbers.
    Usages from Spock, ex.:
    pt6_base 99 [1 3]
    pt6_base 99 1 3
    """

    param_def = [
        ['val1', Type.Float, None, 'value 1'],
        ['numb_list', [['pos', Type.Float, None, 'value']], None,
         'List of values'],
    ]

    def run(self, *args, **kwargs):
        self.data = args


class pt10_base(Macro):
    """Macro with a list of numbers followed by a boolean parameter.
    Usages from Spock, ex.:
    pt10_base [1] True
    pt10_base 1 True
    """

    param_def = [

        ['numb_list', [['pos', Type.Float, None, 'value']], None,
         'List of values'],
        ['val1', Type.Boolean, None, 'value 1'],
    ]

    def run(self, *args, **kwargs):
        self.data = args




import poolunittest
import pool
import HTMLTestRunner


def go(test_output):
    test_output_file = file(test_output, 'wb')
    runner = HTMLTestRunner.HTMLTestRunner(
        stream=test_output_file,
        title='Device Pool Unit Test',
        description='Device Pool Unit Test package',
        verbosity=2)

    super_suite = poolunittest.PoolTestSuite()
    super_suite.addTest(pool.suite())
    runner.run(super_suite)

if __name__ == "__main__":
    test_output = "test_result.html"
    go(test_output)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__docformat__ = 'restructuredtext'

SERVER_NAME = "Sardana"


def prepare_sardana(util):
    from . import pool
    from . import macroserver
    pool.prepare_pool(util)
    macroserver.prepare_macroserver(util)


def main_sardana(args=None, start_time=None, mode=None):
    from .core import util
    # pass server name so the scripts generated with setuptools work on Windows
    return util.run(prepare_sardana, args=args, start_time=start_time,
                    mode=mode, name=SERVER_NAME)

run = main_sardana


def main():
    import datetime
    run(start_time=datetime.datetime.now())



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the sequence macro module"""

__all__ = ["sequence"]

import xml.dom.minidom

from sardana.macroserver.macro import *

TAG_MACRO = 'macro'
TAG_PARAM = 'param'
TAG_PARAMREPEAT = 'paramrepeat'
TAG_REPEAT = 'repeat'
TAG_PARAMS = 'params'
TAG_SEQUENCE = 'sequence'

ATTRIBUTE_NAME = 'name'
ATTRIBUTE_VALUE = 'value'
ATTRIBUTE_INDEX = 'nr'
ATTRIBUTE_DESCRIPTION = 'description'
ATTRIBUTE_DEFVALUE = 'defvalue'
ATTRIBUTE_TYPE = 'type'
ATTRIBUTE_ALLOWEDHOOKS = 'allowedHooks'
ATTRIBUTE_HASPARAMS = 'hasParams'
ATTRIBUTE_MIN = 'min'


class sequence(Macro):
    """This macro executes a sequence of macros. As a parameter
    it receives a string which is a xml structure. These macros which allow
    hooks can nest another sequence (xml structure). In such a case,
    this macro is executed recursively."""

    param_def = [
        ['xml',   Type.String,   None, 'Xml string representing a sequence']
    ]

    def run(self, *pars):
        xmlDoc = xml.dom.minidom.parseString(pars[0])
        macros = self.parseXml(xmlDoc)
        for macro in macros:
            self.runMacro(macro)
#            self.pausePoint()

    def parseXml(self, xmlDoc):
        macros = []
        sequenceElement = xmlDoc.getElementsByTagName(TAG_SEQUENCE)[0]
        childElement = sequenceElement.firstChild
        while childElement:
            if childElement.localName == TAG_MACRO:
                params, hookElement = self.parseMacro(childElement)
                macro = self.createMacro(params)
                if hookElement is not None:
                    hook = self.createExecMacroHook([self.__class__.__name__,
                                                     hookElement.toxml()])
                    macro.hooks = [hook]
                macros.append(macro)
            childElement = childElement.nextSibling
        return macros

    def parseMacro(self, xmlElement):
        name = str(xmlElement.getAttribute(ATTRIBUTE_NAME))
        params = (name,)
        hookElement = None
        childElement = xmlElement.firstChild
        while childElement:
            if childElement.localName == TAG_PARAM:
                params += self.parseParam(childElement)
            elif childElement.localName == TAG_PARAMREPEAT:
                params += self.parseParamRepeat(childElement)
            elif childElement.localName == TAG_SEQUENCE:
                hookElement = childElement
            childElement = childElement.nextSibling
        return (params, hookElement)

    def parseParam(self, xmlElement):
        return (str(xmlElement.getAttribute(ATTRIBUTE_VALUE)),)

    def parseParamRepeat(self, xmlElement):
        params = ()
        childElement = xmlElement.firstChild
        while childElement:
            if childElement.localName == TAG_REPEAT:
                params += self.parseRepeat(childElement)
            childElement = childElement.nextSibling
        return params

    def parseRepeat(self, xmlElement):
        params = ()
        childElement = xmlElement.firstChild
        while childElement:
            if childElement.localName == TAG_PARAM:
                params += self.parseParam(childElement)
            elif childElement.localName == TAG_PARAMREPEAT:
                params += self.parseParamRepeat(childElement)
            childElement = childElement.nextSibling
        return params



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of the macroserver base types for
macros"""

__all__ = ["Integer", "Float", "Boolean", "String", "User", "Filename",
           "File", "Macro", "MacroLibrary", "Env", "Motor", "MotorParam",
           "MotorGroup", "ExpChannel", "MeasurementGroup", "ComChannel",
           "IORegister", "Controller", "Instrument", "ControllerClass",
           "TriggerGate"]

__docformat__ = 'restructuredtext'

from sardana import INTERFACES
from sardana.macroserver.msparameter import ParamType, AttrParamType, \
    ElementParamInterface

# Basic types


class Any(ParamType):
    type_class = lambda p: p


class Integer(ParamType):
    type_class = int


class Float(ParamType):
    type_class = float


class Boolean(ParamType):
    type_class = bool

    def getObj(self, str_repr):
        str_repr = str_repr.lower()
        if str_repr in ['true', '1']:
            value = True
        elif str_repr in ['false', '0']:
            value = False
        else:
            raise ValueError('{0} is not a boolean'.format(str_repr))
        return value


class String(ParamType):
    type_class = str


class User(ParamType):
    type_class = str


class Filename(ParamType):
    type_class = str


class File(ParamType):
    type_class = str

    def __init__(self, macro_server, name):
        ParamType.__init__(self, macro_server, name)
        self.filename = None
        # self.data is supposed to be an array.array object
        self.data = None

    def set(self, filename, data):
        self.filename = filename
        self.data = data


class JSON(ParamType):

    def getObj(self, str_repr):
        import json
        return json.loads(str_repr)


class Env(ParamType):
    type_class = str


class TangoDevice(ParamType):

    def getObj(self, str_repr):
        import PyTango
        return PyTango.DeviceProxy(str_repr)


class Device(ParamType):

    def getObj(self, str_repr):
        import taurus
        return taurus.Device(str_repr)

# Hardware types


class MotorParam(AttrParamType):
    """Class designed to represent a motor parameter name. Usual values
    are acceleration,deceleration,velocity,backlash,steps_per_unit,etc"""

    def __init__(self, macro_server, name):
        AttrParamType.__init__(self, macro_server, name)
        self.attr_item_list = ["Acceleration", "Backlash", "Base_rate", "Step_per_unit",
                               "Deceleration", "Velocity", "Offset"]
        self.non_attr_item_list = ["Controller"]

    def getItemList(self):
        return self.non_attr_item_list + self.attr_item_list

    def getAttrItemList(self):
        return self.attr_item_list

    def getNonAttrItemList(self):
        return self.non_attr_item_list


def __build_base_types():
    for sardana_type, info in list(INTERFACES.items()):
        _, doc = info

        class _I(ElementParamInterface):
            __doc__ = doc
            __name__ = sardana_type
        globals()[sardana_type] = _I

__build_base_types()



import re
import os
import tempfile
import numpy as np
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from taurus.external.unittest import TestCase, main
from taurus.external import qt
from taurus.external.qt import Qt
from sardana.spock.ipython_01_00.genutils import _create_config_file, \
    from_name_to_tango
from sardana.sardanacustomsettings import UNITTEST_DOOR_NAME
from sardana.taurus.qt.qtgui.extra_sardana.qtspock import QtSpockWidget, \
    get_spock_profile_dir

if qt.PYQT4:
    from PyQt4.QtTest import QTest
elif qt.PYQT5:
    from PyQt5.QtTest import QTest
elif qt.PYSIDE:
    from PySide.QtTest import QTest


app = Qt.QApplication.instance()
if not app:
    app = Qt.QApplication([])


def waitFor(predicate, timeout):
    """Process events until predicate is true or timeout seconds passed

    This seems to not handle the destruction of widget correctly, use
    waitForLoop in such cases.
    """
    if predicate():
        return True

    timer = Qt.QElapsedTimer()
    timer.start()

    while not timer.hasExpired(timeout):
        QTest.qWait(min(100, timeout - timer.elapsed()))
        if predicate():
            return True

    return predicate()


def waitForLoop(predicate, timeout):
    """Run event loop and periodically check for predicate"""
    if predicate():
        return True

    timer = Qt.QElapsedTimer()
    timer.start()

    loop = Qt.QEventLoop()

    while not timer.hasExpired(timeout):
        Qt.QTimer.singleShot(min(100, timeout - timer.elapsed()), loop.quit)
        loop.exec_()
        if predicate():
            return True

    return predicate()


class QtSpockBaseTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        # Use setUpClass instead of setUp because starting QtSpock
        # takes a relatively long time.
        cls.test_ipython_dir = tempfile.mkdtemp()
        cls._create_profile()
        os.environ["IPYTHONDIR"] = cls.test_ipython_dir
        cls._isDestroyed = False

    @classmethod
    def _create_profile(cls):
        cls.test_spockdoor_dir = os.path.join(
            cls.test_ipython_dir, "profile_spockdoor")
        os.mkdir(cls.test_spockdoor_dir)
        _create_config_file(cls.test_spockdoor_dir, UNITTEST_DOOR_NAME)

    @classmethod
    def _handle_destroyed(cls):
        cls._isDestroyed = True

    @classmethod
    def tearDownClass(cls):
        cls.widget.shutdown_kernel()
        cls.widget.setAttribute(Qt.Qt.WA_DeleteOnClose)
        cls.widget.destroyed.connect(cls._handle_destroyed)
        cls.widget.close()
        cls.widget = None
        waitForLoop(lambda: cls._isDestroyed, 5000)


class QtSpockTestCase(QtSpockBaseTestCase):
    @classmethod
    def setUpClass(cls):
        super(QtSpockTestCase, cls).setUpClass()
        cls.widget = QtSpockWidget(use_model_from_profile=True)
        cls.widget.start_kernel()
        cls.widget.show()


class CorrectProfileOutputMixin(object):
    def test_spock_banner(self):
        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(r"^Spock \d\.\d", text, re.MULTILINE)
            return len(matches) == 1
        self.assertTrue(waitFor(predicate, 10000))

    def test_spock_prompt(self):
        full_name, name, alias = from_name_to_tango(UNITTEST_DOOR_NAME)
        alias = alias or name

        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(
                r"^{}.*\[(\d)\]".format(alias),
                text, re.MULTILINE)
            return len(matches) == 1 and matches[0] == "1"
        self.assertTrue(waitFor(predicate, 10000))


class CorrectProfileTestCase(QtSpockTestCase, CorrectProfileOutputMixin):
    def test_get_value(self):
        msg_id = self.widget.blocking_client.execute(
            "a = arange(3)", silent=True)
        self.widget.blocking_client.get_shell_msg(msg_id)
        self.assertTrue(
            np.array_equal(self.widget.get_value("a"), np.arange(3)))

    def test_find_spock_profile(self):
        self.assertEqual(
            get_spock_profile_dir("spockdoor"), self.test_spockdoor_dir)

    def test_is_valid_spock_profile(self):
        self.assertTrue(self.widget.kernel_manager.is_valid_spock_profile)


class ProfileErrorOutputMixin(object):
    def test_ipython_banner(self):
        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(r"^IPython \d\.\d", text, re.MULTILINE)
            return len(matches) == 1
        self.assertTrue(waitFor(predicate, 5000))

    def test_ipython_prompt(self):
        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(r"^In.*\[(\d)\]", text, re.MULTILINE)
            return len(matches) == 1 and matches[0] == "1"
        self.assertTrue(waitFor(predicate, 5000))

    def test_profile_error_info(self):
        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(r"^Spock profile error", text, re.MULTILINE)
            return len(matches) == 1
        self.assertTrue(waitFor(predicate, 5000))


class MissingProfileTestCase(QtSpockTestCase, ProfileErrorOutputMixin):
    @classmethod
    def setUpClass(cls):
        cls.test_ipython_dir = tempfile.mkdtemp()
        os.environ["IPYTHONDIR"] = cls.test_ipython_dir
        cls.widget = QtSpockWidget(use_model_from_profile=True)
        cls.widget.start_kernel()
        cls.widget.show()
        cls._isDestroyed = False

    def test_is_valid_spock_profile(self):
        self.assertTrue(not self.widget.kernel_manager.is_valid_spock_profile)


class CorrectProfileAfterRestartTestCase(
        MissingProfileTestCase, CorrectProfileOutputMixin):
    @classmethod
    def setUpClass(cls):
        super(CorrectProfileAfterRestartTestCase, cls).setUpClass()

        # Wait until startup finished
        def predicate():
            text = cls.widget._control.toPlainText()
            matches = re.findall(r"\[1\]: $", text, re.MULTILINE)
            return len(matches) == 1
        assert waitFor(predicate, 5000)

        cls._create_profile()

        # Restart kernel
        def acceptDialog():
            topLevelWidgets = Qt.QApplication.topLevelWidgets()
            dialog = None
            for widget in topLevelWidgets:
                if isinstance(widget, Qt.QMessageBox):
                    dialog = widget
            dialog.button(Qt.QMessageBox.Yes).click()

        Qt.QTimer.singleShot(10, acceptDialog)
        cls.widget.restart_kernel("Restart?")

    def test_is_valid_spock_profile(self):
        self.assertTrue(self.widget.kernel_manager.is_valid_spock_profile)


class ProfileErrorAfterRestartTestCase(
        QtSpockTestCase, ProfileErrorOutputMixin):
    @classmethod
    def setUpClass(cls):
        super(ProfileErrorAfterRestartTestCase, cls).setUpClass()

        # Wait until startup finished
        def predicate():
            text = cls.widget._control.toPlainText()
            matches = re.findall(r"\[1\]: $", text, re.MULTILINE)
            return len(matches) == 1
        assert waitFor(predicate, 5000)

        # "Update" profile
        config_file = os.path.join(
            cls.test_spockdoor_dir, "ipython_config.py")
        with open(config_file) as f:
            config = f.read()
        config = re.sub(
            r"^# spock_creation_version = \d\.\d.*",
            "# spock_creation_version = 9.9.9-gamma",
            config, flags=re.MULTILINE)
        with open(config_file, 'w') as f:
            f.write(config)

        # Restart kernel
        cls.widget.kernel_manager.restart_kernel()

    def test_is_valid_spock_profile(self):
        self.assertTrue(not self.widget.kernel_manager.is_valid_spock_profile)


class QtSpockNoModelTestCase(QtSpockBaseTestCase, ProfileErrorOutputMixin):
    @classmethod
    def setUpClass(cls):
        super(QtSpockNoModelTestCase, cls).setUpClass()
        cls.widget = QtSpockWidget()
        cls.widget.start_kernel()
        cls.widget.show()

    def test_is_valid_spock_profile(self):
        self.assertTrue(not self.widget.kernel_manager.is_valid_spock_profile)

    def test_profile_error_info(self):
        def predicate():
            text = self.widget._control.toPlainText()
            matches = re.findall(r"^No door selected", text, re.MULTILINE)
            return len(matches) == 1
        self.assertTrue(waitFor(predicate, 5000))


class QtSpockModelTestCase(QtSpockBaseTestCase, CorrectProfileOutputMixin):
    @classmethod
    def setUpClass(cls):
        super(QtSpockModelTestCase, cls).setUpClass()
        cls.widget = QtSpockWidget()
        cls.widget.setModel(UNITTEST_DOOR_NAME)
        cls.widget.show()

    def test_is_valid_spock_profile(self):
        self.assertTrue(self.widget.kernel_manager.is_valid_spock_profile)

    def test_setModel_twice_no_restart(self):
        with patch.object(
                self.widget.kernel_manager, "restart_kernel", spec=True) as m:
            self.widget.setModel(UNITTEST_DOOR_NAME)
            self.assertTrue(not m.called)


class QtSpockModelAfterRestartTestCase(
        QtSpockBaseTestCase, CorrectProfileOutputMixin):
    @classmethod
    def setUpClass(cls):
        super(QtSpockModelAfterRestartTestCase, cls).setUpClass()
        cls.widget = QtSpockWidget()
        cls.widget.start_kernel()
        cls.widget.show()

        # Wait until startup finished
        def predicate():
            text = cls.widget._control.toPlainText()
            matches = re.findall(r"\[1\]: $", text, re.MULTILINE)
            return len(matches) == 1
        assert waitFor(predicate, 5000)

        cls.widget.setModel(UNITTEST_DOOR_NAME)

    def test_is_valid_spock_profile(self):
        self.assertTrue(self.widget.kernel_manager.is_valid_spock_profile)


class QtSpockNoModelAfterRestartTestCase(QtSpockNoModelTestCase):
    @classmethod
    def setUpClass(cls):
        # Do only call the original base class constructor
        super(QtSpockNoModelTestCase, cls).setUpClass()
        cls.widget = QtSpockWidget()
        cls.widget.setModel(UNITTEST_DOOR_NAME)
        cls.widget.show()

        # Wait until startup finished
        def predicate():
            text = cls.widget._control.toPlainText()
            matches = re.findall(r"\[1\]: $", text, re.MULTILINE)
            return len(matches) == 1
        assert waitFor(predicate, 5000)

        cls.widget.setModel("")


class QtSpockModelMissingProfileTestCase(
        QtSpockTestCase, ProfileErrorOutputMixin):
    @classmethod
    def setUpClass(cls):
        cls.test_ipython_dir = tempfile.mkdtemp()
        os.environ["IPYTHONDIR"] = cls.test_ipython_dir
        cls.widget = QtSpockWidget()
        cls.widget.setModel(UNITTEST_DOOR_NAME)
        cls.widget.show()
        cls._isDestroyed = False

    def test_is_valid_spock_profile(self):
        self.assertTrue(not self.widget.kernel_manager.is_valid_spock_profile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["OneDExpChannel", "OneDExpChannelClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import DevFailed, DevVoid, DevString, DevState, AttrQuality, \
    Except, READ, SCALAR, READ_WRITE

from taurus.core.util.log import DebugIt

from sardana import State, DataFormat, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool.controller import OneDController, MaxDimSize, Type
from sardana.tango.core.util import to_tango_type_format, exception_str

from sardana.tango.pool.PoolDevice import PoolTimerableDevice, \
    PoolTimerableDeviceClass


class OneDExpChannel(PoolTimerableDevice):

    def __init__(self, dclass, name):
        PoolTimerableDevice.__init__(self, dclass, name)
        self._first_read_cache = False

    def init(self, name):
        PoolTimerableDevice.init(self, name)

    def get_oned(self):
        return self.element

    def set_oned(self, oned):
        self.element = oned

    oned = property(get_oned, set_oned)

    @DebugIt()
    def delete_device(self):
        PoolTimerableDevice.delete_device(self)
        oned = self.oned
        if oned is not None:
            oned.remove_listener(self.on_oned_changed)

    @DebugIt()
    def init_device(self):
        PoolTimerableDevice.init_device(self)
        oned = self.oned
        if oned is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.oned = oned = \
                self.pool.create_element(type="OneDExpChannel",
                                         name=name, full_name=full_name, id=self.Id, axis=self.Axis,
                                         ctrl_id=self.Ctrl_id)
            if self.instrument is not None:
                oned.set_instrument(self.instrument)
        oned.add_listener(self.on_oned_changed)

        # force a state read to initialize the state attribute
        #state = ct.state
        self.set_state(DevState.ON)

    def on_oned_changed(self, event_source, event_type, event_value):
        try:
            self._on_oned_changed(event_source, event_type, event_value)
        except not DevFailed:
            msg = 'Error occurred "on_oned_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.motor.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_oned_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name.lower()
        name = name.replace('_', '')  # for integration_time events
        attr_name = name

        try:
            attr = self.get_attribute_by_name(attr_name)
        except DevFailed:
            return

        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        value, w_value, error = None, None, None

        if name == "state":
            value = self.calculate_tango_state(event_value)
        elif name == "status":
            value = self.calculate_tango_status(event_value)
        elif name == "valuebuffer":
            value = self._encode_value_chunk(event_value)
            self._first_read_cache = True
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                else:
                    value = event_value.value
                timestamp = event_value.timestamp
            else:
                value = event_value
            if name == "timer" and value is None:
                value = "None"
            elif name == "value":
                w_value = event_source.get_value_attribute().w_value
                state = self.oned.get_state()
                if state == State.Moving:
                    quality = AttrQuality.ATTR_CHANGING

        self.set_attribute(attr, value=value, w_value=w_value,
                           timestamp=timestamp, quality=quality,
                           priority=priority, error=error, synch=False)

    def always_executed_hook(self):
        #state = to_tango_state(self.oned.get_state(cache=False))
        pass

    def read_attr_hardware(self, data):
        pass

    def get_dynamic_attributes(self):
        cache_built = hasattr(self, "_dynamic_attributes_cache")

        std_attrs, dyn_attrs = \
            PoolTimerableDevice.get_dynamic_attributes(self)

        if not cache_built:
            # For value attribute, listen to what the controller says for data
            # type (between long and float) and length
            value = std_attrs.get('value')
            if value is not None:
                _, data_info, attr_info = value
                ttype, _ = to_tango_type_format(attr_info.dtype)
                data_info[0][0] = ttype
                shape = attr_info.maxdimsize
                data_info[0][3] = shape[0]
        return std_attrs, dyn_attrs

    def initialize_dynamic_attributes(self):
        attrs = PoolTimerableDevice.initialize_dynamic_attributes(self)

        non_detect_evts = "valuebuffer",

        for attr_name in non_detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, False)

    def read_Value(self, attr):
        oned = self.oned
        # TODO: decide if we force the controller developers to store the
        # last acquired value in the controllers or we always will use
        # cache. This is due to the fact that the clients (MS) read the value
        # after the acquisition had finished.
        use_cache = oned.is_in_operation() and not self.Force_HW_Read
        # For the moment we just check if we recently receive ValueBuffer.
        # event. In this case, we use cache and clean the flag
        # so the cached value will be returned only at the first readout
        # after the acquisition. This is a workaround for the count executed
        # by the MacroServer e.g. step scans or ct which read the value after
        # the acquisition.
        if not use_cache and self._first_read_cache:
            use_cache = True
            self._first_read_cache = False
        value = oned.get_value(cache=use_cache, propagate=0)
        if value.error:
            Except.throw_python_exception(*value.exc_info)
        state = oned.get_state(cache=use_cache, propagate=0)
        quality = None
        if state == State.Moving:
            quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=value.value, quality=quality,
                           timestamp=value.timestamp, priority=0)

    def is_Value_allowed(self, req_type):
        if self.get_state() in [DevState.FAULT, DevState.UNKNOWN]:
            return False
        return True

    def read_DataSource(self, attr):
        data_source = self.oned.get_data_source()
        if data_source is None:
            full_name = self.get_full_name()
            # for Taurus 3/4 compatibility
            if not full_name.startswith("tango://"):
                full_name = "tango://{0}".format(full_name)
            data_source = "{0}/value".format(full_name)
        attr.set_value(data_source)

    def Start(self):
        self.oned.start_acquisition()


_DFT_VALUE_INFO = OneDController.standard_axis_attributes['Value']
_DFT_VALUE_MAX_SHAPE = _DFT_VALUE_INFO[MaxDimSize]
_DFT_VALUE_TYPE, _DFT_VALUE_FORMAT = to_tango_type_format(
    _DFT_VALUE_INFO[Type], DataFormat.OneD)


class OneDExpChannelClass(PoolTimerableDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }
    device_property_list.update(PoolTimerableDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'Start':   [[DevVoid, ""], [DevVoid, ""]],
    }
    cmd_list.update(PoolTimerableDeviceClass.cmd_list)

    #    Attribute definitions
    attr_list = {
        'DataSource': [[DevString, SCALAR, READ]],
    }
    attr_list.update(PoolTimerableDeviceClass.attr_list)

    standard_attr_list = {
        'Value': [[_DFT_VALUE_TYPE, _DFT_VALUE_FORMAT, READ,
                   _DFT_VALUE_MAX_SHAPE[0]],
                  {'abs_change': '1.0', }]
    }
    standard_attr_list.update(PoolTimerableDeviceClass.standard_attr_list)

    def _get_class_properties(self):
        ret = PoolTimerableDeviceClass._get_class_properties(self)
        ret['Description'] = "1D device class"
        ret['InheritedFrom'].insert(0, 'PoolTimerableDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
dooroutput.py:
"""

import taurus.core
from taurus.external.qt import Qt



class DoorOutput(Qt.QPlainTextEdit):
    """Widget used for displaying changes of door's attributes: Output, Info, Warning and Error."""

    def __init__(self, parent=None):
        Qt.QPlainTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))
        self.stopAction = Qt.QAction("Stop scrolling", self)
        self.stopAction.setCheckable(True)
        self.stopAction.setChecked(False)
        self._isStopped = False
        self.showDebug = Qt.QAction("Show debug details", self)
        self.showDebug.setCheckable(True)
        self.showDebug.setChecked(False)
        self._isDebugging = False

    def onDoorOutputChanged(self, output):
        """call on output attribute changed"""
        txt = "<font color=\"Black\">"
        if output is None:
            return
        for i, line in enumerate(output):
            if i > 0:
                txt += "<br/>"
            txt += line.replace(' ', '&nbsp;')
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorInfoChanged(self, info):
        """call on info attribute changed"""
        txt = "<font color=\"Blue\">"
        if info is None:
            return
        for i, line in enumerate(info):
            if i > 0:
                txt += "<br/>"
            txt += line.replace(' ', '&nbsp;')
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorWarningChanged(self, warning):
        """call on warning attribute changed"""
        txt = "<font color=\"Orange\">"
        if warning is None:
            return
        for i, line in enumerate(warning):
            if i > 0:
                txt += "<br/>"
            txt += line.replace(' ', '&nbsp;')
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorErrorChanged(self, error):
        """call on error attribute changed"""
        txt = "<font color=\"Red\">"
        if error is None:
            return
        for i, line in enumerate(error):
            if i > 0:
                txt += "<br/>"
            txt += line.replace(' ', '&nbsp;')
        txt += "</font>"
        self.appendHtmlText(txt)

    def onDoorDebugChanged(self, debug):
        """call on debug attribute changed"""
        txt = "<font color=\"Grey\">"
        if self._isDebugging:
            if debug is None:
                return
            for i, line in enumerate(debug):
                if i > 0:
                    txt += "<br/>"
                txt += line.replace(' ', '&nbsp;')
            txt += "</font>"
            self.appendHtmlText(txt)
            if not self._isStopped:
                self.moveCursor(Qt.QTextCursor.End)

    def appendHtmlText(self, text):
        self.appendHtml(text)
        if not self._isStopped:
            self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        menu.addAction(self.stopAction)
        menu.addAction(self.showDebug)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        clearAction.triggered.connect(self.clear)
        self.stopAction.toggled.connect(self.stopScrolling)
        self.showDebug.toggled.connect(self.showDebugDetails)
        menu.exec_(event.globalPos())

    def stopScrolling(self, stop):
        self._isStopped = stop

    def showDebugDetails(self, debug):
        self._isDebugging = debug


class DoorDebug(Qt.QPlainTextEdit):
    """Deprecated. Do not use"""

    """Widget used for displaying changes of door's Debug attribute."""

    def __init__(self, parent=None):
        Qt.QPlainTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))
        self.stopAction = Qt.QAction("Stop scrolling", self)
        self.stopAction.setCheckable(True)
        self.stopAction.setChecked(False)
        self._isStopped = False

        from taurus.core.util.log import warning

        msg = ("DoorDebug is deprecated since version 3.0.3. "
               "Use DoorOutput 'Show debug details' feature instead.")
        warning(msg)

    def onDoorDebugChanged(self, debug):
        """call on debug attribute changed"""
        if debug is None:
            return
        for line in debug:
            self.appendPlainText(line)

        if not self._isStopped:
            self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        menu.addAction(self.stopAction)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        clearAction.triggered.connect(self.clear)
        self.stopAction.toggled.connect(self.stopScrolling)
        menu.exec_(event.globalPos())

    def stopScrolling(self, stop):
        self._isStopped = stop


class DoorResult(Qt.QPlainTextEdit):
    """Widget used for displaying changes of door's Result attribute."""

    def __init__(self, parent=None):
        Qt.QPlainTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))

    def onDoorResultChanged(self, result):
        """call on result attribute changed"""
        if result is None:
            return
        for line in result:
            self.appendPlainText(line)
        self.moveCursor(Qt.QTextCursor.End)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        clearAction = Qt.QAction("Clear", menu)
        menu.addAction(clearAction)
        if not len(self.toPlainText()):
            clearAction.setEnabled(False)

        clearAction.triggered.connect(self.clear)
        menu.exec_(event.globalPos())


if __name__ == "__main__":
    import sys
    import taurus
    from taurus.core.util.argparse import get_taurus_parser
    from taurus.qt.qtgui.application import TaurusApplication

    parser = get_taurus_parser()
    app = TaurusApplication(sys.argv, cmd_line_parser=parser)
    args = app.get_command_line_args()

    doorOutput = DoorOutput()
    if len(args) == 1:
        door = taurus.Device(args[0])
        door.outputUpdated.connect(doorOutput.onDoorOutputChanged)
        door.infoUpdated.connect(doorOutput.onDoorInfoChanged)
        door.warningUpdated.connect(doorOutput.onDoorWarningChanged)
        door.errorUpdated.connect(doorOutput.onDoorErrorChanged)
        door.debugUpdated.connect(doorOutput.onDoorDebugChanged)

    doorOutput.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This file contains the basic pool definitions."""

__all__ = ["ControllerAPI", "AcqTriggerType", "AcqMode", "SynchDomain",
           "SynchParam", "AcqSynch", "AcqSynchType"]

__docformat__ = 'restructuredtext'

from operator import __getitem__
from enum import IntEnum
from taurus.core.util.enumeration import Enumeration
from sardana.taurus.core.tango.sardana import AcqTriggerType, AcqMode

#: A constant defining the controller API version currently supported
ControllerAPI = 1.1

# synchronization domain: Time means that the configuration parameter will be
# expressed in the time domain, Position means the motor position domain and
# Monitor means the count to monitor domain


class SynchEnum(IntEnum):

    @classmethod
    def fromStr(cls, string):
        '''Convert string representation of SynchDomain enum e.g.
        'SynchDomain.Time' to SynchDomain objects. It also works with just
        domain strings like 'Time'. The following expressions are True:

        SynchDomain.fromStr(str(SynchDomain.Time)) == SynchDomain.Time
        SynchDomain.fromStr('Time') == SynchDomain.Time
        '''
        domain = string.split('.')
        if len(domain) == 1:
            return __getitem__(cls, domain[0])
        elif len(domain) == 2:
            return __getitem__(cls, domain[1])
        else:
            raise ValueError('Can not convert %s to SynchDomain' % string)


class SynchDomain(SynchEnum):
    """Enumeration of synchronization domains.

    - Time - describes the synchronization in time domain
    - Position - describes the synchronization in position domain
    - Monitor - not used at the moment but foreseen for synchronization on
      monitor

    .. note::
        The SynchDomain class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.
    """
    Time = 0
    Position = 1
    Monitor = 2
#     - Default - the controller selects the most appropriate domain:
#       for active events the precedence should be first Position and then
#       Time
#       for passive events the precedence should be first Time and then
#       Position
#    Default = 3


class SynchParam(SynchEnum):
    """Enumeration of synchronization description group parameters.

    - Delay - initial delay (relative to the synchronization start)
    - Total - total interval
    - Active - active interval (part of the total interval)
    - Repeats - number of repetitions within the group
    - Initial - initial point (absolute)

    .. note::
        The SynchParam class has been included in Sardana
        on a provisional basis. Backwards incompatible changes
        (up to and including removal of the class) may occur if
        deemed necessary by the core developers.

    """
    Delay = 0
    Total = 1
    Active = 2
    Repeats = 3
    Initial = 4


AcqSynchType = Enumeration("AcqSynchType", ["Trigger", "Gate", "Start"])
AcqSynchType.__doc__ = \
    """Enumeration of synchronization types.

    Options:

    - Trigger - Start each acquisition (experimental channel will decide on
      itself when to end, based on integration time / monitor count)
    - Gate - Start and end each acquisition
    - Start - Start only the first acquisition (experimental channel will
      drive the acquisition based on integration time / monitor count, latency
      time and number of repetitions)

    .. todo:: convert to python enums, but having in mind problems with
             JSON serialization: https://bugs.python.org/issue18264
    """


class AcqSynch(IntEnum):
    """Enumeration of synchronization options.

    Uses software/hardware naming to refer to internal (software
    synchronizer) or external (hardware synchronization device)
    synchronization modes. See :obj:`~sardana.pool.pooldefs.AcqSynchType`
    to get more details about the synchronization type e.g. trigger, gate or
    start.
    """
    SoftwareTrigger = 0
    """Internal (software) trigger
    
    .. image:: /_static/acqsynch_softtrig.png
    """
    HardwareTrigger = 1
    """External (hardware) trigger
    
    .. image:: /_static/acqsynch_hardtrig.png
    """
    SoftwareGate = 2
    """Internal (software) gate - not implemented
    """
    HardwareGate = 3
    """External (hardware) gate
    
    .. image:: /_static/acqsynch_hardgate.png
    """
    SoftwareStart = 4
    """
    Internal (software) start (triggers just the first acquisition)
    
    .. image:: /_static/acqsynch_softstart.png
    """
    HardwareStart = 5
    """External (hardware) start (triggers just the first acquisition)
    
    .. image:: /_static/acqsynch_hardstart.png
    """

    @classmethod
    def from_synch_type(self, software, synch_type):
        """Helper obtain AcqSynch from information about software/hardware
        nature of synchronization element and AcqSynchType
        """
        if synch_type is AcqSynchType.Trigger:
            if software:
                return AcqSynch.SoftwareTrigger
            else:
                return AcqSynch.HardwareTrigger
        elif synch_type is AcqSynchType.Gate:
            if software:
                return AcqSynch.SoftwareGate
            else:
                return AcqSynch.HardwareGate
        elif synch_type is AcqSynchType.Start:
            if software:
                return AcqSynch.SoftwareStart
            else:
                return AcqSynch.HardwareStart
        else:
            raise ValueError("Unable to determine AcqSynch from %s" %
                             synch_type)



##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import random

from sardana.sardanathreadpool import get_thread_pool
from sardana.util.thread import CountLatch
from unittest import TestCase


def job(i, duration):
    time.sleep(duration)


class CountLatchTestCase(TestCase):

    def test_countlatch(self):
        pool = get_thread_pool()
        latch = CountLatch()
        for i in range(10):
            latch.count_up()
            pool.add(job, latch.count_down, i, random.uniform(0.1, 1))
        latch.wait()
        msg = "some workers are still busy"
        self.assertEqual(pool.getNumOfBusyWorkers(), 0, msg)
        msg = "jobs queue is not empty"
        self.assertEqual(pool.qsize, 0, msg)



import PyTango
import CounterTimerController
import time


class Vct6Controller(CounterTimerController.CounterTimerController):
    "This class is the Tango Sardana CounterTimer controller for the VCT6"

    ctrl_extra_attributes = {'PyCT_extra_1': {'Type': 'PyTango.DevDouble', 'R/W Type': 'PyTango.READ_WRITE'},
                             'PyCT_extra_2': {'Type': 'PyTango.DevLong', 'R/W Type': 'PyTango.READ'},
                             'PyCT_extra_3': {'Type': 'PyTango.DevBoolean', 'R/W Type': 'PyTango.READ'}}

    class_prop = {'CtrlDevName': {'Type': 'PyTango.DevString',
                                  'Description': 'The ctrl simulator Tango device name'}}

    MaxDevice = 6

    def __init__(self, inst, props):
        CounterTimerController.CounterTimerController.__init__(
            self, inst, props)
        print "PYTHON -> Vct6Controller ctor for instance", inst
#	raise NameError,"Ouuups"

        self.ct_name = "Vct6Controller/" + self.inst_name
        self.simu_ctrl = None
        self.simu_ctrl = PyTango.DeviceProxy(self.CtrlDevName)
        self.started = False

        self.dft_PyCT_extra_1 = 88.99
        self.dft_PyCT_extra_2 = 33
        self.dft_PyCT_extra_3 = True

        self.PyCT_extra_1 = []
        self.PyCT_extra_2 = []
        self.PyCT_extra_3 = []

        try:
            self.simu_ctrl.ping()
        except:
            self.simu_ctrl = None
            raise

    def AddDevice(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In AddDevice method for index", ind
        self.PyCT_extra_1.append(self.dft_PyCT_extra_1)
        self.PyCT_extra_2.append(self.dft_PyCT_extra_2)
        self.PyCT_extra_3.append(self.dft_PyCT_extra_3)
#        raise RuntimeError,"Hola la la"

    def DeleteDevice(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In DeleteDevice method for index", ind

    def StateOne(self, ind):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In StateOne
        # method for index",ind
        if self.simu_ctrl is not None:
            if self.started:
                now = time.time()
                delta_t = now - self.start_time
                print "delta_t =", delta_t
                if delta_t > 2.0:
                    for index in self.wantedCT:
                        self.simu_ctrl.command_inout("Stop", index)
                    self.started = False
            sta = self.simu_ctrl.command_inout("GetCounterState", ind)
            # print "State in controller =",sta
            tup = (sta, "Status error string from controller")
        else:
            raise RuntimeError, "Ctrl Tango's proxy null!!!"
        return tup

    def PreReadAll(self):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In PreReadAll
        # method"
        pass

    def PreReadOne(self, ind):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In PreReadOne
        # method for index",ind
        pass

    def ReadAll(self):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In ReadAll
        # method"
        pass

    def ReadOne(self, ind):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In ReadOne method
        # for index",ind
        pass
        if self.simu_ctrl is not None:
            return self.simu_ctrl.command_inout("GetCounterValue", ind)
        else:
            raise RuntimeError, "Ctrl Tango's proxy null!!!"

    def AbortOne(self, ind):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": In AbortOne method for index", ind
        if self.simu_ctrl is not None:
            self.simu_ctrl.command_inout("Stop", ind)
            self.started = False
        else:
            raise RuntimeError, "Ctrl Tango's proxy null!!!"

    def PreStartAllCT(self):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In PreStartAllCT
        # method"
        self.wantedCT = []

    def StartOneCT(self, ind):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In StartOneCT
        # method for index",ind
        self.wantedCT.append(ind)

    def StartAllCT(self):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In StartAllCT
        # method"
        for index in self.wantedCT:
            self.simu_ctrl.command_inout("Start", index)
        self.started = True
        self.start_time = time.time()

    def LoadOne(self, ind, value):
        # print "PYTHON -> Vct6Controller/",self.inst_name,": In LoadOne method
        # for index",ind," with value",value
        if self.simu_ctrl is not None:
            self.simu_ctrl.command_inout("Clear", ind)
        else:
            raise RuntimeError, "Ctrl Tango's proxy null!!!"

    def GetExtraAttributePar(self, ind, name):
        if name == "PyCT_extra_1":
            return self.PyCT_extra_1[ind]
        if name == "PyCT_extra_2":
            return self.PyCT_extra_2[ind]
        return self.PyCT_extra_3[ind]

    def SetExtraAttributePar(self, ind, name, value):
        if name == "PyCT_extra_1":
            self.PyCT_extra_1[ind] = value
        elif name == "PyCT_extra_2":
            self.PyCT_extra_2[ind] = value
        else:
            self.PyCT_extra_3[ind] = value

    def SendToCtrl(self, in_data):
        print "Received value =", in_data
        return "Adios"

    def __del__(self):
        print "PYTHON -> Vct6Controller/", self.inst_name, ": Aarrrrrg, I am dying"


if __name__ == "__main__":
    obj = Vct6Controller('test')
#    obj.AddDevice(2)
#    obj.DeleteDevice(2)
